
"""
$(TYPEDEF)

trilinearform assembly pattern type
"""
abstract type APT_TrilinearForm <: AssemblyPatternType end

function Base.show(io::IO, ::Type{APT_TrilinearForm})
    print(io, "TrilinearForm")
end

"""
````
function TrilinearForm(
    T::Type{<:Real},
    FES::Array{FESpace,1},          
    operators::Array{DataType,1},
    action::AbstractAction;
    regions::Array{Int,1} = [0])
````

Creates a TrilinearForm assembly pattern with the given FESpaces, operators and action etc.
"""
function TrilinearForm(
    T::Type{<:Real},
    AT::Type{<:AssemblyType},
    FES::Array{<:FESpace{Tv,Ti},1},
    operators::Array{DataType,1},
    action::AbstractAction;
    name = "TLF",
    regions::Array{Int,1} = [0]) where {Tv,Ti}
    @assert length(operators) == 3
    @assert length(FES) == 3
    return  AssemblyPattern{APT_TrilinearForm, T, AT}(name,FES,operators,action,[1,2],regions)
end


"""
````
assemble!(
    assemble!(
    A::AbstractArray{T,2},                  # target matrix
    AP::AssemblyPattern{APT,T,AT},          # TrilinearForm pattern
    fixedFE::Array{<:FEVectorBlock,1};      # coefficients for fixed argument
    fixed_argument = [1],                   # position of fixed argument (as an array)
    transposed_assembly::Bool = false,      # transpose result?
    factor = 1)                             # factor that is multiplied
    where {APT <: APT_TrilinearForm, T, AT}
````

Assembly of a TrilinearForm AP into given two-dimensional AbstractArray (e.g. a FEMatrixBlock).
Here, one argument (specified by fixed_argument) is fixed by the given coefficients in fixedFE[1]. Note, that the action is
(currently) always applied to the first and second argument.
"""
function assemble!(
    A::AbstractArray{T,2},
    AP::AssemblyPattern{APT,T,AT},
    fixedFE::Array{<:FEVectorBlock,1};
    fixed_arguments = [1],
    transposed_assembly::Bool = false,
    skip_preps::Bool = false,
    factor = 1) where {APT <: APT_TrilinearForm, T <: Real, AT <: AssemblyType}

    # prepare assembly
    FE = AP.FES
    if !skip_preps
        prepare_assembly!(AP)
    end
    AM::AssemblyManager{T} = AP.AM
    xItemVolumes::Array{T,1} = FE[1].xgrid[GridComponentVolumes4AssemblyType(AT)]
    xItemRegions::GridRegionTypes{Int32} = FE[1].xgrid[GridComponentRegions4AssemblyType(AT)]
    nitems = length(xItemVolumes)

    # prepare action
    action = AP.action
    @assert AP.apply_action_to == [1,2] "Currently it is only allowed to apply to the action to arguments [1,2] (consider changing the order of the arguments and work with fixed_argument instead)"
    if typeof(action) <: NoAction
        @warn "TrilinearForm does not work with no action."
    end
    action_resultdim::Int = action.argsizes[1]
    action_input::Array{T,1} = zeros(T,action.argsizes[2]) # heap for action input
    action_result::Array{T,1} = zeros(T,action_resultdim) # heap for action output

    fixed_argument = fixed_arguments[1]
    if AP.regions != [0]
        @logmsg MoreInfo "Assembling $(AP.name) for given $(fixedFE[1].name) (id=$fixed_argument) into matrix ($AT in regions = $(AP.regions))"
    else
        @logmsg MoreInfo "Assembling $(AP.name) for given $(fixedFE[1].name) (id=$fixed_argument) into matrix ($AT)"
    end
    @assert length(fixed_arguments) == 1 "Matrix assembly for Trilinearform needs exactly one fixed argument"
    @debug AP

    # loop over items
    maxdofitems::Array{Int,1} = get_maxdofitems(AM)
    indexmap = CartesianIndices(zeros(Int, (maxdofitems[1],maxdofitems[2],maxdofitems[3])))
    basisevaler::Array{FEBasisEvaluator,1} = [get_basisevaler(AM, 1, 1),get_basisevaler(AM, 2, 1),get_basisevaler(AM, 3, 1)]
    basisvals_testfunction::Union{SharedCValView{T},Array{T,3}} = basisevaler[3].cvals
    basisxref::Array{Array{T,1},1} = basisevaler[2].xref
    nonfixed_ids::Array{Int,1} = setdiff([1,2,3], fixed_argument)
    coeffs::Array{T,1} = zeros(T,get_maxndofs(AM,fixed_argument))
    dofs2::Array{Int,1} = zeros(Int,get_maxndofs(AM,nonfixed_ids[1]))
    dofs3::Array{Int,1} = zeros(Int,get_maxndofs(AM,nonfixed_ids[2]))
    localmatrix::Array{T,2} = zeros(T,length(dofs2),length(dofs3))
    evalfixedFE::Array{T,1} = zeros(T,size(basisevaler[fixed_argument].cvals,1)) # evaluation of argument 1
    ndofs4item::Array{Int, 1} = [0,0,0]
    dofitem::Array{Int,1} = [0,0,0]
    offsets::Array{Int,1} = [0, size(basisevaler[1].cvals,1), size(basisevaler[1].cvals,1) + size(basisevaler[2].cvals,1)]
    weights::Array{T,1} = get_qweights(AM) # somehow this saves A LOT allocations
    temp::T = 0 # some temporary variable

    regions::Array{Int,1} = AP.regions
    allitems::Bool = (regions == [0])
    nregions::Int = length(regions)
    
    loop_allocations = @allocated for item = 1 : nitems
    for r = 1 : nregions
    # check if item region is in regions
    if allitems || xItemRegions[item] == regions[r]

        # update assembly manager (also updates necessary basisevaler)
        update_assembly!(AM, item)
        weights = get_qweights(AM)

        # loop over associated dofitems
        # di, dj == 2 is only performed if one of the operators jumps
        for di in indexmap
            dofitem[1] = AM.dofitems[1][di[1]]
            dofitem[2] = AM.dofitems[2][di[2]]
            dofitem[3] = AM.dofitems[3][di[3]]

            if dofitem[1] > 0 && dofitem[2] > 0 && dofitem[3] > 0
                # get number of dofs on this dofitem
                ndofs4item[1] = get_ndofs(AM, 1, di[1])
                ndofs4item[2] = get_ndofs(AM, 2, di[2])
                ndofs4item[3] = get_ndofs(AM, 3, di[3])

                # update FEbasisevaler
                basisevaler[1] = get_basisevaler(AM, 1, di[1])::FEBasisEvaluator
                basisevaler[2] = get_basisevaler(AM, 2, di[2])::FEBasisEvaluator
                basisevaler[3] = get_basisevaler(AM, 3, di[3])::FEBasisEvaluator

                # update action on dofitem
                basisxref = basisevaler[2].xref
                update_action!(action, basisevaler[2], dofitem[2], item, regions[r])

                # update coeffs of fixed and dofs of free arguments
                get_coeffs!(coeffs, fixedFE[1], AM, fixed_argument, di[fixed_argument])
                get_dofs!(dofs2, AM, nonfixed_ids[1], di[nonfixed_ids[1]])
                get_dofs!(dofs3, AM, nonfixed_ids[2], di[nonfixed_ids[2]])

                if fixed_argument in [1,2]
                    basisvals_testfunction = basisevaler[nonfixed_ids[2]].cvals
                    for i in eachindex(weights)
    
                        # evaluate fixed argument into action
                        fill!(action_input, 0)
                        eval_febe!(action_input, basisevaler[fixed_argument], coeffs, i, offsets[fixed_argument])
                        
                        for dof_i = 1 : ndofs4item[nonfixed_ids[1]]
                            # apply action to fixed argument and first non-fixed argument
                            eval_febe!(action_input, basisevaler[nonfixed_ids[1]], dof_i, i, offsets[nonfixed_ids[1]])
                            apply_action!(action_result, action_input, action, i, basisxref[i])
            
                            for dof_j = 1 : ndofs4item[nonfixed_ids[2]]
                                temp = 0
                                for k = 1 : action_resultdim
                                    temp += action_result[k] * basisvals_testfunction[k,dof_j,i]
                                end
                                localmatrix[dof_i,dof_j] += temp * weights[i] 
                            end
                        end 
                    end
                else # fixed argument is the last one
                    for i in eachindex(weights)
    
                        # evaluate fixed argument into separate vector
                        fill!(evalfixedFE, 0)
                        eval_febe!(evalfixedFE, basisevaler[fixed_argument], coeffs, i, 0)
                        
                        for dof_i = 1 : ndofs4item[nonfixed_ids[1]]
                            # apply action to fixed argument and first non-fixed argument
                            eval_febe!(action_input, basisevaler[nonfixed_ids[1]], dof_i, i, 0)
                            
                            for dof_j = 1 : ndofs4item[nonfixed_ids[2]]
                                eval_febe!(action_input, basisevaler[nonfixed_ids[2]], dof_j, i, offsets[2])
                                apply_action!(action_result, action_input, action, i, basisxref[i])

                                temp = 0
                                for k = 1 : action_resultdim
                                    temp += action_result[k] * evalfixedFE[k]
                                end
                                localmatrix[dof_i,dof_j] += temp * weights[i]
                            end
                        end 
                    end
                end 
        
                # copy localmatrix into global matrix
                temp = AM.coeff4dofitem[fixed_argument][di[fixed_argument]] * AM.coeff4dofitem[nonfixed_ids[2]][di[nonfixed_ids[2]]] * AM.coeff4dofitem[nonfixed_ids[1]][di[nonfixed_ids[1]]] * xItemVolumes[item] * factor
                for dof_i = 1 : ndofs4item[nonfixed_ids[1]], dof_j = 1 : ndofs4item[nonfixed_ids[2]]
                   # if localmatrix[dof_i,dof_j] != 0
                        if transposed_assembly == true
                            _addnz(A,dofs3[dof_j],dofs2[dof_i],localmatrix[dof_i,dof_j], temp)
                        else
                            _addnz(A,dofs2[dof_i],dofs3[dof_j],localmatrix[dof_i,dof_j], temp)
                        end
                   # end
                end
                fill!(localmatrix,0.0)
            end
        end 
        break; # region for loop
    end # if in region    
    end # region for loop
    end # item for loop
    AP.last_allocations = loop_allocations
    return nothing
end


"""
````
assemble!(
    assemble!(
    b::AbstractVector,                      # target vector
    AP::AssemblyPattern{APT,T,AT},          # coefficients for fixed arguments
    fixedFE::Array{<:FEVectorBlock,1};      # TrilinearForm pattern
    factor = 1)                             # factor that is multiplied
    where {APT <: APT_TrilinearForm, T, AT}
````

Assembly of a TrilinearForm AP into given one-dimensional AbstractArray (e.g. a FEVectorBlock).
Here, the first two arguments are fixed by the given coefficients in FE1 and FE2.
"""
function assemble!(
    b::AbstractVector,
    AP::AssemblyPattern{APT,T,AT},
    fixedFE::Array{<:FEVectorBlock,1};
    skip_preps::Bool = false,
    factor::Real = 1,
    offset::Int = 0) where {APT <: APT_TrilinearForm, T <: Real, AT <: AssemblyType}

    # prepare assembly
    FE1 = fixedFE[1]
    FE2 = fixedFE[2]
    FE = AP.FES
    if !skip_preps
        prepare_assembly!(AP)
    end
    AM::AssemblyManager{T} = AP.AM
    xItemVolumes::Array{T,1} = FE[1].xgrid[GridComponentVolumes4AssemblyType(AT)]
    xItemRegions::GridRegionTypes{Int32} = FE[1].xgrid[GridComponentRegions4AssemblyType(AT)]
    nitems = length(xItemVolumes)

    # prepare action
    action = AP.action
    if typeof(action) <: NoAction
        @warn "TrilinearForm does not work with no action."
    end
    action_resultdim::Int = action.argsizes[1]
    action_input::Array{T,1} = zeros(T,action.argsizes[2]) # heap for action input
    action_result::Array{T,1} = zeros(T,action_resultdim) # heap for action output

    if AP.regions != [0]
        @logmsg MoreInfo "Assembling $(AP.name) for given $((p->p.name).([FE1,FE2])) into vector ($AT in regions = $(AP.regions)) into vector"
    else
        @logmsg MoreInfo "Assembling $(AP.name) for given $((p->p.name).([FE1,FE2])) into vector ($AT) into vector"
    end
    @debug AP

    # loop over items
    maxdofitems::Array{Int,1} = get_maxdofitems(AM)
    indexmap = CartesianIndices(zeros(Int, (maxdofitems[1],maxdofitems[2],maxdofitems[3])))
    basisevaler::Array{FEBasisEvaluator,1} = [get_basisevaler(AM, 1, 1),get_basisevaler(AM, 2, 1),get_basisevaler(AM, 3, 1)]
    basisvals_testfunction::Union{SharedCValView{T},Array{T,3}} = basisevaler[3].cvals
    basisxref::Array{Array{T,1},1} = basisevaler[2].xref
    fixed_arguments::Array{Int,1} = [1,2]
    nonfixed_id::Int = 3
    coeffs1::Array{T,1} = zeros(T,get_maxndofs(AM,fixed_arguments[1]))
    coeffs2::Array{T,1} = zeros(T,get_maxndofs(AM,fixed_arguments[2]))
    dofs3::Array{Int,1} = zeros(Int,get_maxndofs(AM,nonfixed_id))
    dofitem::Array{Int,1} = [0,0,0]
    offsets::Array{Int,1} = [0, size(basisevaler[1].cvals,1), size(basisevaler[1].cvals,1) + size(basisevaler[2].cvals,1)]
    weights::Array{T,1} = get_qweights(AM) # somehow this saves A LOT allocations
    itemfactor::T = 0 # some temporary variable
    localb::Array{T,1} = zeros(T,length(dofs3))

    regions::Array{Int,1} = AP.regions
    allitems::Bool = (regions == [0])
    nregions::Int = length(regions)
    loop_allocations = @allocated for item = 1 : nitems
    for r = 1 : nregions
    # check if item region is in regions
    if allitems || xItemRegions[item] == regions[r]

        # update assembly manager (also updates necessary basisevaler)
        update_assembly!(AM, item)
        weights = get_qweights(AM)

        # loop over associated dofitems
        # di, dj == 2 is only performed if one of the operators jumps
        for di in indexmap
            dofitem[1] = AM.dofitems[1][di[1]]
            dofitem[2] = AM.dofitems[2][di[2]]
            dofitem[3] = AM.dofitems[3][di[3]]

            if dofitem[1] > 0 && dofitem[2] > 0 && dofitem[3] > 0

                # update FEbasisevaler
                basisevaler[1] = get_basisevaler(AM, 1, di[1])::FEBasisEvaluator
                basisevaler[2] = get_basisevaler(AM, 2, di[2])::FEBasisEvaluator
                basisevaler[3] = get_basisevaler(AM, 3, di[3])::FEBasisEvaluator
                basisvals_testfunction = basisevaler[3].cvals
                basisxref = basisevaler[2].xref

                # update action on dofitem
                update_action!(action, basisevaler[2], dofitem[2], item, regions[r])

                # update dofs of free arguments
                get_coeffs!(coeffs1, FE1, AM, fixed_arguments[1], di[fixed_arguments[1]])
                coeffs1 .*= AM.coeff4dofitem[fixed_arguments[1]][di[fixed_arguments[1]]]
                get_coeffs!(coeffs2, FE2, AM, fixed_arguments[2], di[fixed_arguments[2]])
                coeffs2 .*= AM.coeff4dofitem[fixed_arguments[2]][di[fixed_arguments[2]]]
                get_dofs!(dofs3, AM, nonfixed_id, di[nonfixed_id])

                for i in eachindex(weights)
                    # evaluate first and second component
                    fill!(action_input, 0)
                    eval_febe!(action_input, basisevaler[fixed_arguments[1]], coeffs1, i)
                    eval_febe!(action_input, basisevaler[fixed_arguments[2]], coeffs2, i, offsets[2])
        
                    # apply action to FE1 and FE2
                    apply_action!(action_result, action_input, action, i, basisxref[i])
                   
                    # multiply third component
                    for dof_j = 1 : get_ndofs(AM, 3, di[3])
                        temp = 0
                        for k = 1 : action_resultdim
                            temp += action_result[k] * basisvals_testfunction[k,dof_j,i]
                        end
                        localb[dof_j] += temp * weights[i]
                    end 
                end 
        
                itemfactor = factor * xItemVolumes[item] * AM.coeff4dofitem[nonfixed_id][di[nonfixed_id]]
                for dof_i = 1 : get_ndofs(AM, 3, di[3])
                    b[get_dof(AM, 3, di[3], dof_i) + offset] += localb[dof_i] * itemfactor
                end
                fill!(localb, 0.0)
            end
        end 
        break; # region for loop
    end # if in region    
    end # region for loop
    end # item for loop
    AP.last_allocations = loop_allocations
    return nothing
end


## wrapper for FEVectorBlock to avoid use of setindex! functions of FEMAtrixBlock
function assemble!(
    b::FEVectorBlock,
    AP::AssemblyPattern{APT,T,AT},
    fixedFE::Array{<:FEVectorBlock,1};
    skip_preps::Bool = false,
    fixed_arguments = [1,2], # currently the only accepted configuration
    factor::Real = 1) where {APT <: APT_TrilinearForm, T <: Real, AT <: AssemblyType}

    assemble!(b.entries, AP, fixedFE; factor = factor, offset = b.offset, skip_preps = skip_preps)
end